const chainIds = {
  hazama: 1,
  ethereum: 1,
  rinkeby: 4,
  goerli: 5,
  polygon: 137,
  mumbai: 80001,
  astar: 592,
}

exports.SUPPORTED_CHAIN_NAME = ['ethereum', 'goerli', 'polygon', 'mumbai']

exports.WaitingAlchemyPeriod = 120000

exports.AlchemyApiKeys = Object.freeze({
  goerli: 'td3dNUbdRxiWjq40dzROhGQnEs5t_wQM',
  ethereum: '7XkQ3F7K3sdroSJonHDcTYA6cs1i3VYN',
  polygon: 'k81asPDI6lu1XyXhjpEs7Xz7vVTT-1wQ',
  mumbai: 'OHyYiQUHvfY4VSi5CfP1nkt-E08R1jMC',
})

exports.AlchemyNetworks = Object.freeze({
  goerli: 'eth-goerli',
  ethereum: 'eth-mainnet',
  polygon: 'polygon-mainnet',
  mumbai: 'polygon-mumbai',
})

exports.NetworkInfo = Object.freeze({
  hazama: {
    chainId: `0x${chainIds.hazama.toString(16)}`,
    chainName: 'Hazama Network',
    name: 'Hazama',
    rpcUrls: ['https://mainnet.infura.io/v3/9aa3d95b3bc440fa88ea12eaa4456161'],
  },
  ethereum: {
    chainId: `0x${chainIds.ethereum.toString(16)}`,
    chainName: 'Ethereum Network',
    name: 'Ethereum',
    nativeCurrency: {
      name: 'ETH',
      symbol: 'ETH',
      decimals: 18,
    },
    rpcUrls: [
      `https://${exports.AlchemyNetworks.ethereum}.g.alchemy.com/v2/${exports.AlchemyApiKeys.ethereum}`,
    ],
  },
  polygon: {
    chainId: `0x${chainIds.polygon.toString(16)}`,
    chainName: 'Polygon Network',
    name: 'Polygon',
    nativeCurrency: {
      name: 'MATIC',
      symbol: 'MATIC',
      decimals: 18,
    },
    rpcUrls: [
      `https://${exports.AlchemyNetworks.polygon}.g.alchemy.com/v2/${exports.AlchemyApiKeys.polygon}`,
    ],
  },
  astar: {
    chainId: `0x${chainIds.astar.toString(16)}`,
    chainName: 'Astar Network',
    name: 'Astar',
    nativeCurrency: {
      name: 'ASTR',
      symbol: 'ASTR',
      decimals: 18,
    },
    rpcUrls: ['https://rpc.astar.network:8545'],
    blockExplorerUrls: ['https://astar.subscan.io/'],
  },
  goerli: {
    chainId: `0x${chainIds.goerli.toString(16)}`,
    chainName: 'Goerli Network',
    name: 'Goerli',
    nativeCurrency: {
      name: 'ETH',
      symbol: 'ETH',
      decimals: 18,
    },
    rpcUrls: [
      `https://${exports.AlchemyNetworks.goerli}.g.alchemy.com/v2/${exports.AlchemyApiKeys.goerli}`,
    ],
  },
  mumbai: {
    chainId: `0x${chainIds.mumbai.toString(16)}`,
    chainName: 'Mumbai Network',
    name: 'Mumbai',
    nativeCurrency: {
      name: 'MATIC',
      symbol: 'MATIC',
      decimals: 18,
    },
    rpcUrls: [
      `https://${exports.AlchemyNetworks.mumbai}.g.alchemy.com/v2/${exports.AlchemyApiKeys.mumbai}`,
    ],
  },
})

/**
 *
 * @private
 * @param {string} network network
 * @returns {string}
 */
exports.getRpcUrlsByNetWork = (network) => {
  return network.startsWith('http') || network.startsWith('ws')
    ? network
    : exports.NetworkInfo[network]?.rpcUrls?.[0] || ''
}

/**
 *
 * @private
 * @param {string} network network
 * @returns {string}
 */
exports.getAlchemyApiKey = (network) => {
  return exports.AlchemyApiKeys[network] || ''
}

/**
 *
 * @private
 * @param {string} network network
 * @returns {string}
 */
exports.getAlchemyNetwork = (network) => {
  return exports.AlchemyNetworks[network] || ''
}

/**
 *
 * @private
 * @param {string} network network
 * @returns {number | null}
 */
exports.getChainId = (network) => {
  return chainIds?.[network] || null
}

exports.HazamaHost = 'wss://n.hazbase.com'
