const ContractType = require('../constant/contract-type')

function hasLength(val) {
  return val.length > 0
}

function isArray(val) {
  return Array.isArray(val)
}

function isArrayFull(val) {
  return isArray(val) && hasLength(val)
}

function sleep(ms) {
  return new Promise((resolve) => setTimeout(resolve, ms))
}

function rejectNull(arr) {
  return arr.filter((val) => val != null)
}

function objectToLowerCase(object) {
  for (const x in object) {
    const index = x.toLowerCase()
    object[index] = object[x]
    delete object[x]
  }
  return object
}
function generatePath({ walletAddress, projectId, contractAddress }) {
  return `/hazama/${walletAddress}/projects/${projectId}/contracts/${contractAddress}`
}

/*
  Combine asm to execute two templates in a single transaction.
*/
function jointAsm(asm, asm2) {
  try {
    const asmArray = asm.split(/\n/)
    const asm2Array = asm2.split(/\n/)

    asmArray.pop() // delete END
    asm2Array.shift() // delete #asm
    asm2Array.shift() // delete REQUIRE

    return asmArray.concat(asm2Array).join('\n')
  } catch (e) {}

  return asm
}

async function iterate({ lib, path, match }) {
  path = path?.replace(/\/$/, '')

  const iter = await lib.iterate(
    {
      gt: `${path}`,
      lt: `${path}~`,
      match: `^${path}/${match}$`,
    },
    10,
  )

  const ret = []
  let chunk = await iter()
  while (chunk !== null) {
    for (let i = 0; i < chunk.length; i++) {
      const { key, value } = chunk[i]
      ret.push([key, value])
    }
    chunk = await iter()
  }
  return ret
}

function isNull(val) {
  // eslint-disable-next-line valid-typeof
  return val === null
}

function isString(val) {
  return typeof val === 'string'
}

function isStringFull(val) {
  return isString(val) && hasLength(val)
}

function isObject(val) {
  return typeof val === 'object' && !isNull(val)
}

function isObjectFull(val) {
  return isObject(val) && hasLength(Object.keys(val))
}

function isUndefined(val) {
  return typeof val === 'undefined'
}

function isNil(val) {
  return isUndefined(val) || isNull(val)
}

function getContractInfoFromPath(path) {
  const pattern = /\/hazama\/(.*)\/projects\/(.*)\/contracts\/(.*)/gi
  path = path.replace(pattern, '$1_$2_$3')
  path = path.split('_')
  return {
    walletAddress: path?.[0] || null,
    projectId: path?.[1] || null,
    contractId: path?.[2] || null,
  }
}

function getContractName(contractType) {
  for (const key in ContractType) {
    if (ContractType[key] === contractType) return key
  }
  return undefined
}

function encodeQueryData(data) {
  const result = []
  for (const d in data) {
    result.push(encodeURIComponent(d) + '=' + encodeURIComponent(data[d]))
  }
  return result.join('&')
}

module.exports = {
  hasLength,
  isArray,
  isArrayFull,
  sleep,
  rejectNull,
  objectToLowerCase,
  generatePath,
  jointAsm,
  iterate,
  isNull,
  isString,
  isStringFull,
  isObject,
  isObjectFull,
  isUndefined,
  isNil,
  getContractInfoFromPath,
  getContractName,
  encodeQueryData,
}
